/*******************************************************************************
 * Copyright (C) 2024
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
#include "router.h"

// Router module with FIFOs, semaphores, and capacity control for receive FIFO
router::router(sc_module_name name)
: in_fifo(CORE_FIFO_CAPACITY), out_fifo(ROUTER_RECEIVE_FIFO_CAPACITY), sem(1), current_router_id(-1) // default router id
{
    sc_report_handler::set_log_file_name("out/report.log");
    sc_report_handler::set_actions(LOG_NAME, SC_INFO, SC_LOG|SC_DISPLAY);
    for (int i = 0; i < CORE_NO; ++i) {
        //changed name to match receiving function
        sockets[i].register_nb_transport_fw(this, &router::nb_transport_fw); //register the nb_transport fun to each target socket to receive data from ni
    }
       
    SC_THREAD(process_in_fifo); //process to handle data in the input fifo where data from ni has been saved and send them destination router
    SC_THREAD(process_out_fifo); //process to send the data available in output fifo to destination ni
}

// Function to handle incoming transactions, changed function name 
tlm::tlm_sync_enum router::nb_transport_fw(tlm::tlm_generic_payload& trans, tlm::tlm_phase& phase, sc_time& delay) {
    if (phase == tlm::BEGIN_REQ) {
        if (in_fifo.num_free() == 0) {
            log_info("Router "+ std::string(name()) +": Input FIFO full, cannot accept transaction with data: ");
            delay = sc_time(5, SC_NS); // Delay before retry
            
            return tlm::TLM_UPDATED; // Indicate to retry later
        }
        sc_time current_time = sc_time_stamp();
        //manually copy the transaction payload for future use
        my_payload* orig_trans = dynamic_cast<my_payload*>(&trans);
        my_payload* my_trans = new my_payload();
        // Log the time when the ni sent the transaction
        sc_time sent_time = orig_trans->timestamp;
        my_trans->dst_ni = orig_trans->dst_ni;
        my_trans->dst_router = orig_trans->dst_router;
        my_trans->data = orig_trans->data;
        my_trans->src_router = orig_trans->src_router;
        int src_router_id = orig_trans->src_router;
        my_trans->src_ni = orig_trans->src_ni;
        my_trans->update_timestamp(current_time);  // Update the timestamp in the payload
       
        my_trans->set_address(orig_trans->get_address());
        my_trans->set_command(orig_trans->get_command());
        my_trans->set_data_ptr(reinterpret_cast<unsigned char*>(&my_trans->data));
        my_trans->set_data_length(orig_trans->get_data_length());
        my_trans->set_streaming_width(orig_trans->get_streaming_width());
        my_trans->set_byte_enable_ptr(orig_trans->get_byte_enable_ptr());
        my_trans->set_dmi_allowed(orig_trans->is_dmi_allowed());
        my_trans->set_response_status(tlm::TLM_INCOMPLETE_RESPONSE);
        // Push the transaction into the input FIFO by blocking write
        in_fifo.write(my_trans);
        
        log_info("Router " + std::to_string(src_router_id) + ": Received transaction with data " 
                 + " at time: " + current_time.to_string());

        phase = tlm::END_REQ;  // Mark the end of the request phase
        delay = sc_time(10, SC_NS);  // Add some delay
        return tlm::TLM_UPDATED;
    }
    return tlm::TLM_ACCEPTED;
}



// Thread to process in_fifo
void router::process_in_fifo() {  
    my_payload* payload;
    while (true) {
        unsigned int current_router = current_router_id;
        // Wait for data in the input FIFO
        payload = in_fifo.read();  // Blocking read  
        //std::cout<<in_fifo.num_free()<<std::endl;
        // Decode the address to determine the destination router and ni
        unsigned int address = payload->get_address();
        unsigned int destination_router_id = (address >> 16) & 0xF;  // Extract router ID from address
        unsigned int destination_ni_id = (address >> 8) & 0xFF;    // Extract ni ID from address   
        unsigned int src_ni = payload->src_ni;
        std::cout<<" Check: Source NI id from address at source router: "<< src_ni<<std::endl; 
        std::cout<<" Destination NI id from address at source router: "<< destination_ni_id<<std::endl;
        
        ////A pre-check to determine whether the token is available to acquire and use, currently disabled 
        // if(sem_mgr->peek_token(current_router_id, destination_router_id)) {
        //     log_info("Router " + std::to_string(current_router) + " : Token for destination Router: "  
        //             + std::to_string(destination_router_id) + " is currently available to acquire" );
        // }
        // else {
        //     log_info("Router " + std::to_string(current_router) + " : Token for destination Router: "  
        //             + std::to_string(destination_router_id) + " is currently unavailable to acquire" );
        // }
        // Retry mechanism
        bool success = false;
        while (!success) {
            if((current_router_id == destination_router_id)||(src_ni == destination_ni_id)) {
                if (other_out_fifos[destination_router_id]->num_free() > 0) {
                        
                    int dest_fifo_space = other_out_fifos[destination_router_id]->num_free();
                    log_info("Router: " + std::to_string(current_router_id) 
                            + " The PEs belong to same cluster, no need to wait for token ");
                    // If there's space and the semaphore is available, send the payload
                    log_info("Router: " + std::to_string(current_router_id) + " forwarded data with ID: " 
                            + std::to_string(payload->data->id) + " to Router_" + std::to_string(destination_router_id));
                    wait(1, SC_NS);
                    //blocking write to destination router
                    other_out_fifos[destination_router_id]->write(payload);  // Blocking write to the destination router's FIFO
                    success = true;  // Successfully sent the payload, exit the loop                       

                } else {
                    log_info("Router " + std::to_string(current_router_id) + ": Destination router Router_" + std::to_string(destination_router_id) 
                            + "'s out_fifo full. Retrying with data (ID): " + std::to_string(payload->data->id));
                    wait(sc_time(2, SC_NS));  // Wait before retrying
                } 
            } else if (sem_mgr->request_token(current_router_id, destination_router_id))  {
                
                if (other_out_fifos[destination_router_id]->num_free() > 0) {
                    other_sems[destination_router_id]-> wait();
                        
                    int dest_fifo_space = other_out_fifos[destination_router_id]->num_free();
                    // If there's space and the semaphore is available, send the payload
                    log_info("Router: " +std::to_string(current_router_id) + " forwarded data with ID: " 
                            + std::to_string(payload->data->id) + " to Router_" + std::to_string(destination_router_id));
                    sc_core::sc_time delay_T = sc_core::sc_time(((destination_router_id - current_router + ROUTER_NO) % ROUTER_NO) * CHANNEL_DELAY, SC_NS);  
                    std::cout << "Check:: The delay calculated is: " << delay_T << std::endl;
                    sc_core::sc_time recv_time = sc_time_stamp() + delay_T;
                    payload->update_timestamp(recv_time);
                    wait(delay_T);
                    //blocking write to destination router
                    other_out_fifos[destination_router_id]->write(payload);  // Blocking write to the destination router's FIFO
                    success = true;  // Successfully sent the payload, exit the loop                       
                    other_sems[destination_router_id]->post();  // Release the semaphore
                    // Release the semaphore after forwarding
                    sem_mgr->release_token(current_router_id, destination_router_id);

                } else {
                    log_info("Router " + std::to_string(current_router_id) + ": Destination router Router_" + std::to_string(destination_router_id) + "'s out_fifo full. Retrying with data (ID): " + std::to_string(payload->data->id));
                    other_sems[destination_router_id]->post();  // Release the semaphore  
                    sem_mgr->release_token(current_router_id, destination_router_id);  // Release and retry
                    //Retry delay change from 5 to 2 NS
                    wait(sc_time(2, SC_NS));  // Wait before retrying
                }    
                    
            } else {

                //Retry delay change from 5 to 2 NS
                log_info("Router " + std::to_string(current_router_id) + ": Semaphore for Router_" + std::to_string(destination_router_id) + " is unavailable. Retrying with data (ID): " + std::to_string(payload->data->id));
                wait(sc_time(2, SC_NS));  // Wait before retrying
            }      
        }         

    }
}


// Thread to process router_receive_fifo and forward transactions to NIs
void router::process_out_fifo() {
    
    my_payload* payload;
    int prev_src_id = -1;
    sc_core::sc_time prev_timestamp = sc_time(0.1, SC_NS);
    while (true) {
       
        
        // log_info(std::string(name()) + " has: " + std::to_string(out_fifo.num_free())
        //                          + " remaining space in the output fifo" );

        // Wait for data in the output FIFO
        payload = out_fifo.read();  // Blocking read 
        
        // Extract destination NI ID from the address
        unsigned int address = payload->get_address();
        unsigned int destination_ni_id = (address >> 8) & 0xFF; // NI ID
        unsigned int dest_router_id = (address >> 16) & 0xF;  // Extract router ID from address   
        unsigned int src_router = payload->src_router; 
        sc_core::sc_time timestamp = payload->timestamp;
        //std::cout<<"destination NI id from address at destination router:"<< destination_ni_id<<std::endl;



        int delay_T = (dest_router_id -src_router + ROUTER_NO) % ROUTER_NO * CHANNEL_DELAY;   
        // if(timestamp==prev_timestamp){
        //     log_info("Router " + std::to_string(dest_router_id)+ ": Received transaction with data: " + std::to_string(payload->data) 
        //                 + " from Router " + std::to_string(payload->src_router) + " after a delay of:" + std::to_string(delay_T) + " ns" );  
        //         prev_src_id = src_router;     
        //         prev_timestamp = timestamp;   
        // } else {

        // wait(delay_T, SC_NS);  // Simulate delay for receiving the data 

            log_info("Router " + std::to_string(dest_router_id)+ ": Received transaction with data: " + std::to_string(payload->data->id) 
                        + " from Router " + std::to_string(src_router) + " after a delay of:" + std::to_string(delay_T) + " ns" );
            
            prev_src_id = src_router; 
            prev_timestamp = timestamp; 
        //}
             
       

        log_info("Router " + std::to_string(current_router_id) + ": Forwarding transaction with data: " + std::to_string(payload->data->id) 
                    + " to Processing Element " + std::to_string(destination_ni_id));

        // Send the payload to the corresponding NI using non-blocking TLM
            tlm::tlm_phase phase = tlm::BEGIN_REQ;
            sc_time delay = SC_ZERO_TIME;
            tlm::tlm_sync_enum status = core_sockets[destination_ni_id % CORE_NO]->nb_transport_fw(*payload, phase, delay);

            if (status == tlm::TLM_COMPLETED || phase == tlm::END_REQ) {
                log_info("Router " + std::to_string(current_router_id) + ": Network Interface " + std::to_string(destination_ni_id) 
                        + " completed transaction, data ID: " + std::to_string(payload->data->id));

           ////   other_sems[dest_router_id]->post();
                
            } else if (status == tlm::TLM_ACCEPTED) {
                log_info("Router " + std::to_string(current_router_id) + ": Network Interface " + std::to_string(destination_ni_id) 
                        + " accepted transaction, waiting for completion.");
            } else if (status == tlm::TLM_UPDATED) {
                log_info("Router " + std::to_string(current_router_id) + ": Network Interface " + std::to_string(destination_ni_id) 
                        + " updated transaction, waiting for completion.");
            } else {
                log_error("Router " + std::to_string(current_router_id) + ": Unexpected response from Network Interface " + std::to_string(destination_ni_id));
                //delete payload;
            }
                
            }
}
    
//print log messages
void router::log_info(std::string msg){
    SC_REPORT_INFO(LOG_NAME, msg.c_str());
}

void router::log_error(std::string msg){
    SC_REPORT_ERROR(LOG_NAME, msg.c_str());
}